!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: symboll.gs,v 1.7 2008-01-09 19:02:37 stever Exp $
!
! Superclass Hierarchy:
!   SymbolList, Array, SequenceableCollection, Collection, Object.
!
!=========================================================================

removeallmethods SymbolList
removeallclassmethods SymbolList

category: 'For Documentation Installation only'
classmethod: SymbolList
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A SymbolList is an Array whose elements are instances of SymbolDictionary.

 It is used in compilation of GemStone Smalltalk code, in order to resolve
 references to objects by name.  Given a Symbol as a name, the SymbolList
 searches its dictionary elements in order, and the first key that matches the
 given Symbol then resolves to the object that is the value at that key in that
 dictionary.' .
doc documentClassWith: txt.

self description: doc.
%

category: 'Accessing'
method: SymbolList
names

"Returns an Array of Strings containing the names of the receiver's
 SymbolDictionaries.

 This method assumes that each SymbolDictionary in the receiver contains a
 SymbolAssociation whose value is that SymbolDictionary.  If any
 SymbolDictionary does not contain such a SymbolAssociation, it is
 represented in the result Array as '(unnamed Dictionary)'."

| result aDict aSymAssoc|
result:= Array new .
1 to: self size do:[:j|
  aDict := self at: j .
  aDict == nil
    ifTrue:[ result add: '(nil)' ]
    ifFalse:[
      aSymAssoc:= aDict associationsDetect:[:anAssoc | anAssoc value == aDict] 
                        ifNone:[ nil ].
      aSymAssoc == nil
        ifTrue:[ result add: '(unnamed Dictionary)' ]
        ifFalse:[ result add: (aSymAssoc key)]
      ].
  ].
^ result
%
category: 'Formatting'
method: SymbolList
namesReport

"Returns a formatted String describing the position and name of each Dictionary
 in the receiver's symbol list.

 This method assumes that each Dictionary in the symbol list contains an
 Association whose value is that Dictionary.  If any Dictionary does not
 contain such an Association, it is represented in the result String as
 '(unnamed Dictionary)'."

"used by the topaz list dictionaries command"

| arrayOfNames result lf |

arrayOfNames := self names .
result := String new .
lf := Character lf .
1 to: arrayOfNames size do:[:j|
  result add: (j asString).
  result addAll: '  '.
  result addAll: (arrayOfNames at: j) .
  result add: lf.
  ].
^result
%

category: 'Searching'
method: SymbolList
resolveSymbol: aString

"Searches the receiver for an Association whose key is equal to aString, and
 returns that Association.  If no such Association is found, returns nil."

<primitive: 49>

(aString isKindOf: Symbol) ifFalse:[
  ^ self resolveSymbol: aString asSymbol
  ].
aString _validateClasses: #[String, DoubleByteString].
self _primitiveFailed: #resolveSymbol: .
self _uncontinuableError
%

category: 'Searching'
method: SymbolList
objectNamed: aSymbol

"Returns the first object in the receiver that has the given name.  If no object
 is found with the given name, returns nil."

| assn |

assn := self resolveSymbol: aSymbol.
assn == nil ifTrue: [ ^nil ].
^assn value
%

category: 'Updating'
method: SymbolList
createDictionaryNamed: dictName at: anIndex 

"Creates a new SymbolDictionary in the Segment of the receiver.  Adds to the
 dictionary an Association whose key is dictName and whose value is the
 dictionary.  Inserts the dictionary in the receiver at anIndex.

 If anIndex is less than 1, the dictionary is inserted at the front of the list.
 If anIndex is greater than the size of the list, it is inserted at the end of
 the list.  If the receiver already contains a dictionary named dictName, raises
 an error."

| newDict nameSym offset userProfile|

nameSym := dictName asSymbol .
self do:[ :aDict | 
  (aDict names includesIdentical: nameSym) ifTrue:[ 
    aDict _error: #rtErrDuplicateKey args: #[ nameSym ].
    ].
  ].
newDict := SymbolDictionary new name: nameSym; yourself .
newDict assignToSegment: self segment .
(newDict associationAt: nameSym) assignToSegment: self segment .  
offset := anIndex .
offset < 1 ifTrue:[ offset := 1 ].
offset > (self size + 1)
  ifTrue:[ offset := self size + 1 ].

userProfile := System myUserProfile.
self == userProfile symbolList
  ifTrue:[userProfile insertDictionary: newDict at: offset]
  ifFalse:[ self insertObject: newDict at: offset ].
^self
%

category: 'Updating'
method: SymbolList
removeDictionaryNamed: aSymbol 

"Removes the first dictionary found in the receiver that contains an Association
 whose key is aSymbol and whose value is the dictionary.  Returns the removed
 dictionary.

 If no such dictionary is found, raises a KeyNotFound error."
  | result |
result := self removeDictionaryNamed: aSymbol 
     ifAbsent:[ self _error: #rtErrKeyNotFound args: aSymbol . nil ] .
^result
%

category: 'Updating'
method: SymbolList
replaceElementsFrom: aSymbolList

"Removes all elements in the receiver and inserts all elements of aSymbolList in
 it, in the same order as in aSymbolList.
 
 If the argument is not a SymbolList, raises an error."

| newSize |
aSymbolList == self ifTrue:[ ^ self ].
(aSymbolList isKindOf: SymbolList) ifFalse:[ 
  aSymbolList _validateClass: SymbolList .
  ^ self
  ].
newSize := aSymbolList size.
self size: aSymbolList size .
aSymbolList copyFrom: 1 to: newSize into: self startingAt: 1 .
%

category: 'Updating'
method: SymbolList
removeDictionaryNamed: aSymbol ifAbsent: aBlock 

"Removes the first dictionary found in the receiver that contains an Association
 whose key is aSymbol and whose value is the dictionary.  Returns the removed
 dictionary.

 If no such dictionary is found, returns the result of evaluating the
 zero-argument Block aBlock." 

| nameSym userProfile|

userProfile := System myUserProfile.
nameSym := aSymbol asSymbol .
1 to: self size do:[ :j | | aDict |
  aDict := self at: j .
  (aDict names includesIdentical: nameSym) ifTrue:[
    userProfile symbolList == self
      ifTrue:[userProfile removeDictionaryAt: j]
      ifFalse:[self deleteObjectAt: j].
    ^ aDict .
    ].
  ].
^ aBlock value
%

category: 'Searching'
method: SymbolList
symbolResolutionOf: aString

"Searches the receiver for aString.  If aString is found, returns a formatted
 String that describes the position in the receiver of the Dictionary defining
 aString, the name of that Dictionary, and aString.

 Generates an error if aString is not defined in the receiver."

"Example: GsSession currentSession symbolList symbolResolutionOf: #Integer"

| lf result theAssoc aDict aSymbol |

aSymbol := Symbol _existingWithAll: aString.
aSymbol == nil ifTrue:[ self _error: #rtErrSymNotFound args: #[aString] ].

result := String new.
lf := Character lf.
1 to: self size do:[ :j|
  aDict := self at: j .
  aDict == nil ifFalse:[
    (aDict at: aSymbol ifAbsent: [ nil ]) ~~ nil ifTrue:[
      " found aSymbol in this dict "
      result add: lf; add: (j asString); addAll: '  '.
      theAssoc:= aDict associationsDetect: [:anAssoc | anAssoc value == aDict]
                               ifNone: [ nil ] .
      theAssoc == nil
        ifTrue:[ result add: '(Unnamed Dictionary)']
        ifFalse: [ result add: theAssoc key ].
      result add: lf; addAll: '      '; addAll: aSymbol;
                 addAll: '  '; add: ((aDict at: aSymbol) asString).
      ^ result 
      ].
    ].
  ].
^ self _error: #rtErrSymNotFound args: #[ aString ]
%

category: 'Repository Conversion'
method: SymbolList
convertContentsTo5

"Private."

1 to: self size do: [ :i |
  self at: i put: ((self at: i) convertTo5)
  ] .
self convertNativeLanguage
%

category: 'Repository Conversion'
method: SymbolList
convertNativeLanguage

""

| ugAssoc ug langSym |
ugAssoc := self resolveSymbol: #UserGlobals .
ugAssoc ~~ nil ifTrue:[ 
  ug := ugAssoc value .
  langSym := ug at: #NativeLanguage otherwise: nil .
  langSym ~~ nil ifTrue:[
    ug at: #NativeLanguage put: langSym asSymbol
    ].
  ].
%

category: 'Repository Conversion'
method: SymbolList
convRecompileAllClasses

"Recompiles methods in all classes in the receiver, using the receiver
 as the symbol list for recompilation."

1 to: self size do:[:j |
  (self at: j) convRecompileAllClassesWith: self
  ].
%

category: 'Searching'
method: SymbolList
dictionaryAndSymbolOf: anObject

"Returns the Dictionary that names anObject, and also returns the
 name which that Dictionary associates with anObject.  More precisely, this
 returns an Array containing two elements:

 * The Dictionary in the receiver that contains an Association whose
   value is anObject.
 * The Symbol which is that Association's key.

 The receiver is searched in the same order that the compiler searches it.
 (For more information about symbol resolution, see the GemStone Programming
 Guide.)  If anObject is not found in the receiver, returns nil."


1 to: self size do:[ :j | | aDict |
  aDict := self at: j .
  aDict ~~ nil ifTrue:[
    aDict associationsDo:[ :assoc |    
      assoc value == anObject ifTrue:[ ^ #[ aDict, assoc key ] ].
      ].
    ].
  ].
^ nil
%

